#include <blasfeo_d_aux.h>
#include <blasfeo_s_aux.h>
#include <blasfeo_d_blas.h>
#include <blasfeo_s_blas.h>

#include <cblas.h>

#include <benchmark/benchmark.h>

#include <random>
#include <memory>


namespace tmpc :: benchmark
{
    using std::size_t;


    static void randomize(size_t m, size_t n, blasfeo_dmat * A)
    {
        std::random_device rd;  //Will be used to obtain a seed for the random number engine
		std::mt19937 gen(rd()); //Standard mersenne_twister_engine seeded with rd()
		std::uniform_real_distribution<> dis(-1.0, 1.0);	

        for (size_t i = 0; i < m; ++i)
            for (size_t j = 0; j < n; ++j)
                BLASFEO_DMATEL(A, i, j) = dis(gen);
    }


    static void randomize(size_t m, size_t n, double * A)
    {
        std::random_device rd;  //Will be used to obtain a seed for the random number engine
		std::mt19937 gen(rd()); //Standard mersenne_twister_engine seeded with rd()
		std::uniform_real_distribution<> dis(-1.0, 1.0);	

        for (size_t i = 0; i < m; ++i)
            for (size_t j = 0; j < n; ++j)
                A[i + j * m] = dis(gen);
    }


    static auto alignedAlloc(size_t bytes)
    {
        return std::unique_ptr<char[], decltype(&std::free)>(
            reinterpret_cast<char *>(std::aligned_alloc(0x20, bytes)), &std::free);
    }


    static void BM_gemm_blasfeo(::benchmark::State& state)
    {
        size_t const m = state.range(0);
        size_t const n = state.range(1);
        size_t const k = state.range(3);

        blasfeo_dmat A, B, C, D;

        auto mem_A = alignedAlloc(blasfeo_memsize_dmat(k, m));
        auto mem_B = alignedAlloc(blasfeo_memsize_dmat(k, n));
        auto mem_C = alignedAlloc(blasfeo_memsize_dmat(m, n));
        auto mem_D = alignedAlloc(blasfeo_memsize_dmat(m, n));

        blasfeo_create_dmat(k, m, &A, mem_A.get());
        blasfeo_create_dmat(k, n, &B, mem_B.get());
        blasfeo_create_dmat(m, n, &C, mem_C.get());
        blasfeo_create_dmat(m, n, &D, mem_D.get());

        randomize(k, m, &A);
        randomize(k, n, &B);
        randomize(m, n, &C);

        for (auto _ : state)
            blasfeo_dgemm_tn(m, n, k, 
                1.0, 
                &A, 0, 0, 
                &B, 0, 0, 
                1.0, 
                &C, 0, 0, 
                &D, 0, 0);
    }


    static void BM_gemm_cblas(::benchmark::State& state)
    {
        size_t const m = state.range(0);
        size_t const n = state.range(1);
        size_t const k = state.range(3);

        auto A = std::make_unique<double []>(k * m);
        auto B = std::make_unique<double []>(k * n);
        auto C = std::make_unique<double []>(m * n);

        randomize(k, m, A.get());
        randomize(k, n, B.get());
        randomize(m, n, C.get());

        for (auto _ : state)
            cblas_dgemm(CblasColMajor, CblasTrans, CblasNoTrans, m, n, k,
		    1.0, A.get(), k, B.get(), k, 1.0, C.get(), m);

    }

    
    BENCHMARK(BM_gemm_blasfeo)
        ->Args({2, 2, 2})
        ->Args({3, 3, 3})
        ->Args({5, 5, 5})
        ->Args({10, 10, 10})
        ->Args({20, 20, 20})
        ->Args({30, 30, 30});

    BENCHMARK(BM_gemm_cblas)
        ->Args({2, 2, 2})
        ->Args({3, 3, 3})
        ->Args({5, 5, 5})
        ->Args({10, 10, 10})
        ->Args({20, 20, 20})
        ->Args({30, 30, 30});
}
